/******************************************************************************
* To the extent possible under law, Sebastian Roith
* <https://www.roith-unterrichtet.de> has waived all copyright and related or
* neighboring rights to this script. This work is published from Germany.
******************************************************************************/

/**
* Klasse, die Kartenspiel „Mäuse sammeln“ simuliert
*/
function Kartenspiel(
    parentNode,         // DOM-Knoten, als dessen Kind das Spiel gerendert werden soll
    gegner,             // Gegnerart (0 → Mensch, 1 → zufallsziehung durch Computer, 2 → perfekter Computergegner)
    name                // Spielername
) {
    Object.defineProperty(this, 'parentNode', { "enumerable": true, "value": parentNode });
    this.aufgedeckt = [0, 1, -1];
    this.gegner = Number.isFinite(gegner) ? Math.min(Math.floor(Math.abs(gegner)), 2) : 1;
    name && (this.spieler = name);
    this._establish();
    this.restart();
}

// Karten auf Nachziehstapel
Kartenspiel.prototype.cards = [];

// Aufgedeckte Karten
Kartenspiel.prototype.aufgedeckt = [];

// Gegnerart (0 → Mensch, 1 → zufallsziehung durch Computer, 2 → perfekter Computergegner)
Kartenspiel.prototype.gegner = 1;

// Verzögerung, bis der Computer eine Karte nimmt
Kartenspiel.prototype.gegnerDelay = 1333;

// Was soll getan werden, wenn das Spiel zu Ende ist (Parameter: Gewonnen?, eigene Punkte, Gegnerpunkte
Kartenspiel.prototype.ondone = null;

// Name des Spielers
Kartenspiel.prototype.spieler = 'Mensch';

/**
* Ggf. kartenziehung durch Computer
* @return undefined
*/
Kartenspiel.prototype._computer = function() {
    if (!this.dran || !this.gegner)
        return;
    if (!this.aufgedeckt[1]) {
        if (this.aufgedeckt[2])
            this._karteNehmen(true, 2);
        return;
    }
    if (!this.aufgedeckt[2]) {
        this._karteNehmen(true, 1);
        return;
    }
    switch (this.gegner) {
        case 1:
            this._karteNehmen(true, Math.floor(Math.random() * 2) + 1);
            break;
        default:
            this._karteNehmen(true, (this.aufgedeckt[1] > this.aufgedeckt[2]) ? 1 : 2);
            break;
    }
}

/**
* HTML anlegen
* @return undefined
*/
Kartenspiel.prototype._establish = function() {
    this.spieler2 = ScoTools.createHtml('p', this.parentNode, { "class": "spieler" }, this.getState(true));
    this.node = ScoTools.createHtml('div', this.parentNode, { "class": 'maeusesammeln' });
    var list = ScoTools.createHtml('ul', this.node);
    this._slots = [
        ScoTools.createHtml('li', list, { "class": "nachziehstapel" }),
        ScoTools.createHtml('li', list),
        ScoTools.createHtml('li', list)
    ];
    this.spieler1 = ScoTools.createHtml('p', this.parentNode, { "class": "spieler" }, this.getState(false));
}

/**
* Status von Spieler bzw. Gegner ermitteln
* @return String    Statusinformationen
*/
Kartenspiel.prototype.getState = function(
    gegner      // Status von Gegner?
) {
    var s = gegner ? (this.gegner ? 'Computer' : 'Mitspieler(in)') : this.spieler;
    s += ': ' + (gegner ? this.score2 : this.score1) + ' Punkte';
    if (this.isDone()) {
        s += ((gegner ? (this.score2 > this.score1) : (this.score1 > this.score2)) ? ', gewonnen' : ', verloren');
    } else if (gegner == this.dran)
        s += ', ist dran';
    return s;
}

/**
* Die URL zur gewünschten Karte bestimmen
* @return String    URL
*/
Kartenspiel.getKartenURL = function(
    value       // Optional: Kartenwert (Standard: Kartenrückseite)
) {
    return 'cards/' + (Number.isFinite(value) ? (((value < 0) ? 'minus' : 'plus') + Math.abs(value)) : 'kartenrueckseite') + '.png';
}

/**
* Eine Karte auf einen Stapel legen
* @return undefined
*/
Kartenspiel.prototype._addCard = function(
    stapel,     // Index des Stapels, auf den die Karte gelegt werden soll
    value       // Optional: Kartenwert
) {
    var node = stapel ? ScoTools.createHtml('a', null, { "href": "#", "title": "Diese Karte nehmen" }) : ScoTools.createHtml('span');
    if (stapel) {
        var self = this;
        node.addEventListener('click', function(ev) {
            ev.preventDefault();
            ev.stopPropagation();
            self.procClick(stapel);
        });
    }
    ScoTools.createHtml('img', node, { "class": "karte", "src": Kartenspiel.getKartenURL(value), "alt": Number.isFinite(value) ? (((value < 0) ? 'Karte ' : 'Karte +') + value) : "Kartenrückseite" });
    this.aufgedeckt[stapel] = Number.isFinite(value) ? value : 0;
    this._slots[stapel].appendChild(node);
}

/**
* Unterste Karte von einem Stapel entfernen
* @return HTMLImage     img-Element der entfernten Karte
*/
Kartenspiel.prototype._removeCard = function(
    stapel      // Index des Stapels, von dem die Karte geneommen werden soll
) {
    var node;
    if (node = this._slots[stapel].firstChild) {
        node.parentNode.removeChild(node);
        this.aufgedeckt[stapel] = 0;
        return node.firstChild;
    }
    return null;
}

// Punktestände
Kartenspiel.prototype.score1 = 0;
Kartenspiel.prototype.score2 = 0;

// Ist gerade Spieler 2 dran?
Kartenspiel.prototype.dran = false;

/**
* Punktestand aktualisieren
* @return undefined
*/
Kartenspiel.prototype._displayScore = function(
    spieler2        // Soll der Punktestand des zweiten Spielers aktualisiert werden?
) {
    this.spieler1.textContent = this.getState(false);
    this.spieler2.textContent = this.getState(true);
    if (this.dran) {
        this.spieler1.className = 'spieler';
        this.spieler2.className = 'spieler dran';
    } else {
        this.spieler1.className = 'spieler dran';
        this.spieler2.className = 'spieler';
    }
}

/**
* Wenn möglich: Karte vom Nachziehstapel nachlegen
* @return undefined
*/
Kartenspiel.prototype._nachziehen = function(
    stapel      // Auf welchen Stapel soll die nachgezogene Karte gelegt werden?
) {
    if (!this.cards.length)
        return;
    this._addCard(stapel, this.cards.pop());
    if (this.cards.length < 3)
        this._removeCard(0);
}

/**
* Einem Spieler eine aufgedeckte Karte zuweisen
* @return undefined
*/
Kartenspiel.prototype._karteNehmen = function(
    anderer,        // Nimmt Spieler 2 eine Karte?
    stapel          // Index des Stapels, von dem die Karte genommen wird
) {
    if ((this.dran != anderer) || !this.aufgedeckt[stapel])
        return;
    this.dran = !this.dran;
    if (anderer)
        this.score2 += this.aufgedeckt[stapel];
    else
        this.score1 += this.aufgedeckt[stapel];
    this._removeCard(stapel);
    this._displayScore(anderer);
    this._nachziehen(stapel);
    this.next();
}

/**
* Ermittelt, ob das Spiel zuende ist
* @return Boolean   Zu Ende?
*/
Kartenspiel.prototype.isDone = function() {
    return !this.aufgedeckt.reduce(function(a, b) { return Math.abs(a) + Math.abs(b);});
}

/**
* Bei Bedarf dem Computer eine Karte ziehen lassen oder Melden, dass das Spiel vorbei ist
* @return undefined
*/
Kartenspiel.prototype.next = function() {
    if (this.isDone() && (typeof this.ondone == 'function'))
        this.ondone.call(this, this.score1 > this.score2, this.score1, this.score2);

    // Anstoßen, dass Computer eine Karte zieht
    if (!this.dran || !this.gegner)
        return;
    var self = this;
    setTimeout(function() { self._computer(); }, this.gegnerDelay);
}

/**
* Auf Anklicken einer Karte reagieren
* @return undefined
*/
Kartenspiel.prototype.procClick = function(
    stapel          // Index des angeklickten Stapels
) {
    if (this.dran && this.gegner)
        return;
    this._karteNehmen(this.dran, stapel);
}

/**
* Spiel neu starten
* @return undefined
*/
Kartenspiel.prototype.restart = function() {
    this.cards = [1, 2, 3, 4, 5, -1, -2, -3, -4, -5].map(function(v) { return { "v": v, "rand": Math.random() }; }).sort(function(a, b) { return a.rand - b.rand; }).map(function(o) { return o.v; });
    this.score1 = 0;
    this.score2 = 0;
    this.dran = false;
    this._displayScore(false);
    this._displayScore(true);
    while (this._removeCard(0)) ;
    while (this._removeCard(1)) ;
    while (this._removeCard(2)) ;
    this._addCard(0);
    this._addCard(0);
    this._addCard(0);
    this._addCard(1, this.cards.pop());
    this._addCard(2, this.cards.pop());
}
